import unittest


class Clock(object):
    def __init__(self) -> None:
        self.reset()

    def __call__(self) -> float:
        return self.now

    def reset(self) -> None:
        self.now = 0

    def increment(self, num: int = 1) -> None:
        self.now += num


clock = Clock()

__all__ = ["unittest", "clock"]



from ratelimit import limits, RateLimitException
from tests import unittest, clock


class TestDecorator(unittest.TestCase):
    @limits(calls=1, period=10, clock=clock)
    def increment(self) -> None:
        """
        Increment the counter at most once every 10 seconds.
        """
        self.count += 1

    @limits(calls=1, period=10, clock=clock, raise_on_limit=False)
    def increment_no_exception(self) -> None:
        """
        Increment the counter at most once every 10 seconds, but w/o rasing an
        exception when reaching limit.
        """
        self.count += 1

    def setUp(self) -> None:
        self.count = 0
        clock.increment(10)

    def test_increment(self) -> None:
        self.increment()
        self.assertEqual(self.count, 1)

    def test_exception(self) -> None:
        self.increment()
        self.assertRaises(RateLimitException, self.increment)

    def test_reset(self) -> None:
        self.increment()
        clock.increment(10)

        self.increment()
        self.assertEqual(self.count, 2)

    def test_no_exception(self) -> None:
        self.increment_no_exception()
        self.increment_no_exception()

        self.assertEqual(self.count, 1)



"""
Rate limit utilty functions.
"""

import time
from .types import ClockCallable


def now() -> ClockCallable:
    """
    Use monotonic time if available, otherwise fall back to the system clock.

    :return: Time function.
    :rtype: function
    """
    if hasattr(time, "monotonic"):
        return time.monotonic
    return time.time



from setuptools import setup, find_packages

import ratelimit


with open("README.rst", "r", encoding="utf-8") as fh:
    long_description = fh.read()


setup(
    name="ratelimit-extended",
    version=ratelimit.__version__,
    description="API rate limit decorator",
    long_description=long_description,
    author="Ömer Faruk Yığın",
    author_email="omert1122@gmail.com",
    url="https://github.com/omert11/ratelimit",
    license="MIT",
    packages=find_packages(),
    install_requires=[],
    keywords=["ratelimit", "api", "decorator"],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python",
        "Topic :: Software Development",
    ],
    include_package_data=True,
    zip_safe=False,
)



"""
Function decorator for rate limiting.

This module provides a function decorator that can be used to wrap a function
such that it will raise an exception if the number of calls to that function
exceeds a maximum within a specified time window.

For examples and full documentation see the README at
https://github.com/tomasbasham/ratelimit
"""

from ratelimit.decorators import RateLimitDecorator, sleep_and_retry
from ratelimit.exception import RateLimitException

limits = RateLimitDecorator
rate_limited = RateLimitDecorator  # For backwards compatibility

__all__ = ["RateLimitException", "limits", "rate_limited", "sleep_and_retry"]

__version__ = "1.0.2"



"""
Rate limit public interface.

This module includes the decorator used to rate limit function invocations.
Additionally this module includes a naive retry strategy to be used in
conjunction with the rate limit decorator.
"""

from functools import wraps
import logging
from math import floor

import time
import sys
import threading
from typing import Any, Callable

from ratelimit.exception import RateLimitException
from ratelimit.utils import now
from .types import ClockCallable


logger = logging.getLogger(__name__)


class RateLimitDecorator(object):
    """
    Rate limit decorator class.
    """

    def __init__(
        self,
        calls: int = 15,
        period: int = 900,
        clock: ClockCallable = now(),
        raise_on_limit: bool = False,
        log_on_limit: bool = True,
        log_message: str | None = None,
    ) -> None:
        """
        Instantiate a RateLimitDecorator with some sensible defaults. By
        default the Twitter rate limiting window is respected (15 calls every
        15 minutes).

        :param int calls: Maximum function invocations allowed within a time period.
        :param float period: An upper bound time period (in seconds) before the rate limit resets.
        :param function clock: An optional function retuning the current time.
        :param bool raise_on_limit: A boolean allowing the caller to avoiding rasing an exception.
        """  # noqa: E501
        self.clamped_calls = max(1, min(sys.maxsize, floor(calls)))
        self.period = period
        self.clock = clock
        self.raise_on_limit = raise_on_limit
        self.log_on_limit = log_on_limit
        self.log_message = log_message

        # Initialise the decorator state.
        self.last_reset = clock()
        self.num_calls = 0

        # Add thread safety.
        self.lock = threading.RLock()

    def __call__(self, func: Callable) -> Callable:
        """
        Return a wrapped function that prevents further function invocations if
        previously called within a specified period of time.

        :param function func: The function to decorate.
        :return: Decorated function.
        :rtype: function
        """

        @wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            """
            Extend the behaviour of the decorated function, forwarding function
            invocations previously called no sooner than a specified period of
            time. The decorator will raise an exception if the function cannot
            be called so the caller may implement a retry strategy such as an
            exponential backoff.

            :param args: non-keyword variable length argument list to the decorated function.
            :param kargs: keyworded variable length argument list to the decorated function.
            :raises: RateLimitException
            """  # noqa: E501
            with self.lock:
                period_remaining = self.__period_remaining()

                # If the time window has elapsed then reset.
                if period_remaining <= 0:
                    self.num_calls = 0
                    self.last_reset = self.clock()

                # Increase the number of attempts to call the function.
                self.num_calls += 1

                # If the number of attempts to call the function exceeds the
                # maximum then raise an exception.
                if self.num_calls > self.clamped_calls:
                    if self.raise_on_limit:
                        raise RateLimitException("too many calls", period_remaining)
                    elif self.log_on_limit:
                        logger.warning(
                            self.log_message
                            or "Rate limit exceeded for %s function." % func.__name__,
                        )
                    else:
                        return

            return func(*args, **kwargs)

        return wrapper

    def __period_remaining(self) -> float:
        """
        Return the period remaining for the current rate limit window.

        :return: The remaing period.
        :rtype: float
        """
        elapsed = self.clock() - self.last_reset
        return self.period - elapsed


def sleep_and_retry(func: Callable) -> Callable:
    """
    Return a wrapped function that rescues rate limit exceptions, sleeping the
    current thread until rate limit resets.

    :param function func: The function to decorate.
    :return: Decorated function.
    :rtype: function
    """

    @wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        """
        Call the rate limited function. If the function raises a rate limit
        exception sleep for the remaing time period and retry the function.

        :param args: non-keyword variable length argument list to the decorated function.
        :param kargs: keyworded variable length argument list to the decorated function.
        """  # noqa: E501
        while True:
            try:
                return func(*args, **kwargs)
            except RateLimitException as exception:
                time.sleep(exception.period_remaining)

    return wrapper
